// Window definitions **************************

#include <string.h>
#include <stdio.h>

#include <window.h>
#include <app.h>
#include <main.h>
#include <stringc.h>
#include <wstring.h>

Window::Window(Object *Parent,char *aTitle,char *aName) : Object(Parent)
{
	//grab title
	wTitle = WS (aTitle).dup ();
	Flags=0;
	if (aName)
	{
		Name=new char [strlen (aName) + 1];
		strcpy (Name, aName);
	}
	else
		Name = NULL;

	hWnd=0;
	//set defaults
	Style=/*WS_OVERLAPPEDWINDOW*/ 0;
	x = CW_USEDEFAULT;
	y = 0;
	w = CW_USEDEFAULT;
	h = 0;

	OldStyle = 0;
	Icon=-1;
	// scroll bar stuff
	// sensible defalults ?

	xMax=GetSystemMetrics(/*SM_CXFULLSCREEN*/ SM_CXSCREEN);
	yMax=GetSystemMetrics(/*SM_CYFULLSCREEN*/ SM_CYSCREEN);
	xStep=yStep=1;
	xPos=yPos=0;
}

wchar_t *Window::GetClassName()
{
	return App::wName;
}

/* The Object can be deleted in two ways
1) Windows sends a WM_CLOSE and then a WM_DESTROY message which does 'delete this'
2) The application does 'delete Object' at any time
*/

Window::~Window()
{
	//free memory
	if (wTitle) delete[] wTitle;
	if (Name) delete[] Name;

	// no more messages, not even destroy
	SetWindowLong(hWnd,GWL_WNDPROC,(long) NullProc);
	SetPointer( hWnd, NULL );

	// only destroy if option 2) above
	if (!(Flags & W_DESTROYED)) DestroyWindow();
}

// MDI will change this
void Window::DestroyWindow()
{
	::DestroyWindow(hWnd);
}

BOOL Window::Register()
{
  WNDCLASS wc;

  if ( !GetClassInfo( 0, GetClassName(), &wc) &&
		 !GetClassInfo(App::hInstance, GetClassName(),&wc) )
	{
		wc.style = /* CS_DBLCLKS */ 0; //CS_HREDRAW | CS_VREDRAW;
		wc.lpfnWndProc   = InitialisationProc;
		wc.cbClsExtra    = 0;
		// Reserve extra bytes for each instance of the window;
		// we will use these bytes to store a pointer to the C++
		// (MainWindow) object corresponding to the window.
		// the size of a 'this' pointer depends on the memory model.
		wc.cbWndExtra    = sizeof( Window * );

		wc.hInstance     = App::hInstance;
		wc.hIcon         = Icon>=0 ? LoadIcon( App::hInstance, MAKEINTRESOURCE(Icon))  : NULL;
		wc.hCursor       = LoadCursor( NULL, IDC_ARROW );
		wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);

		wc.lpszMenuName  = NULL;
		wc.lpszClassName = GetClassName();

		return RegisterClass( &wc );
	}
	return TRUE;
}


Window *Window::CreationWindow=NULL;

BOOL Window::Create()
{
	if (!Register()) return FALSE;
	
	if (Style & (WS_HSCROLL | WS_VSCROLL)) Flags |= W_SCROLLBARS;
	CreationWindow=this;
	hWnd=CreateWindow(GetClassName(),wTitle,Style,x,y,w,h,ParentHWnd,
		0,App::hInstance,(LPSTR) this );
	if (!hWnd)	return FALSE;

	if (!SetupWindow()) 
	{
		DestroyWindow();
		return FALSE;
	}

	return TRUE;
}

void Window::Destroy() {}

// could use Window event table for this stuff
BOOL Window::WndProc(TMSG &Msg)
{
	// handle internal actions
	switch (Msg.Msg)
	{
		case WM_CLOSE:
			if (!CanClose()) return TRUE; //returns 0
			break;
		case WM_DESTROY:  // quit if main window
			if (this==App::MainWindow) App::Quit();
			Flags|=W_DESTROYED;
			Destroy();
			// MainWindow is deleted on program exit
			if (this==App::MainWindow) App::Quit();
			else delete this;
			return TRUE;
		case WM_PAINT:
			PAINTSTRUCT  ps;
			BeginPaint(hWnd, (LPPAINTSTRUCT) &ps);
			/*SetWindowOrgEx(ps.hdc,xPos,yPos,NULL);*/
			Paint(ps.hdc,ps.fErase,ps.rcPaint);

			EndPaint(hWnd, (LPPAINTSTRUCT) &ps);
			break;

		case WM_HSCROLL:			HScroll(Msg); break;
		case WM_VSCROLL:			VScroll(Msg); break;
		case WM_SIZE:				WMSize(Msg); break;
 	}
	//allow user to grab any message
	return EV_FIND(Msg);
}

BOOL Window::SetupWindow() { return TRUE; }

void Window::Paint(HDC, BOOL, RECT&) {}

BOOL Window::WMSize(TMSG &Msg)
{
	if (!(Flags & W_SCROLLBARS)) return TRUE;
	
	//int nWidth = LOWORD(Msg.lParam);  /* width of client area              */
	//int nHeight = HIWORD(Msg.lParam); /* height of client area             */
	
	switch (Msg.wParam)
	{
		case SIZE_MAXIMIZED:
		case SIZE_RESTORED:
			SetVirtualExtent(xMax,yMax,xStep,yStep,0,0);
			break;

		case SIZE_MINIMIZED:
		case SIZE_MAXHIDE:
		case SIZE_MAXSHOW:
			break;
	}

	return TRUE;
}


void Window::SetVirtualExtent(long xSize,long ySize,int xSt, int ySt,int xPgSize,int yPgSize)
{
	// prevent reentrance
	static BOOL InFn=FALSE;
	if (InFn) return;
	InFn=TRUE;

	xMax=xSize;
	yMax=ySize;
	xStep=xSt;
	yStep=ySt;

	RECT rc;
	GetClientRect(hWnd,&rc);
	xPageSize=xPgSize ? xPgSize :rc.right;
	yPageSize=yPgSize ? yPgSize :rc.bottom;
	
	SCROLLINFO si;
	si.cbSize=sizeof(si);
	si.fMask=SIF_ALL;
	si.nMin=0;
	si.nMax=xMax-1;
	si.nPage=xPageSize;
	si.nPos=xPos;
	SetScrollInfo(hWnd,SB_HORZ,&si,TRUE);
	si.nMax=yMax-1;
	si.nPage=yPageSize;
	si.nPos=yPos;
	SetScrollInfo(hWnd,SB_VERT,&si,TRUE);
	// scoll bars may have gone
	GetClientRect(hWnd,&rc);
	// resive oversized window
	if (!(Flags & W_OVERSCROLL) && rc.right>=xMax && rc.bottom>=yMax)
	{
		int newW= min( (long) rc.right, xMax );
		int newH= min( (long) rc.bottom, yMax );
		RECT rc2;
		GetWindowRect(hWnd,&rc2);
		SetWindowPos(hWnd,NULL,0,0,newW+rc2.right-rc2.left-rc.right,
				newH+rc2.bottom-rc2.top-rc.bottom,
				SWP_NOMOVE | SWP_NOZORDER);
//		GetClientRect(hWnd,&rc);
	}
	xPageSize=xPgSize ? xPgSize :rc.right;
	yPageSize=yPgSize ? yPgSize :rc.bottom;
	si.fMask=SIF_PAGE;
	si.nPage=xPageSize;
	SetScrollInfo(hWnd,SB_HORZ,&si,TRUE);
	si.nPage=yPageSize;
	SetScrollInfo(hWnd,SB_VERT,&si,TRUE);
	si.fMask=SIF_POS;
	GetScrollInfo(hWnd,SB_HORZ,&si);
	int newxPos=si.nPos;
	GetScrollInfo(hWnd,SB_VERT,&si);
	int newyPos=si.nPos;
	if (newxPos!=xPos || newyPos!=yPos)
	{
		ScrollWindowEx(hWnd,xPos-newxPos, yPos-newyPos, NULL,NULL,NULL,NULL,SW_ERASE);
		xPos=newxPos;
		yPos=newyPos;
		UpdateWindow(hWnd);
	}

	InFn=FALSE;
	Flags |= W_SCROLLBARS;
}

// WM_VSCROLL
BOOL Window::VScroll(TMSG &Msg)
{
	if (!(Flags & W_SCROLLBARS)) return TRUE;

	int nScrollCode = LOWORD(Msg.wParam);   /* scroll bar value     */
	int nPos = LLHW(Msg);   /* scroll box position  */

	long newPos=yPos;
	switch (nScrollCode) {
		case SB_LINEUP:
			newPos-=yStep; break;
		case SB_PAGEUP:
			newPos-=yPageSize; break;
		case SB_LINEDOWN:
			newPos+=yStep; break;
		case SB_PAGEDOWN:
			newPos+=yPageSize; break;
		case SB_THUMBTRACK:
		case SB_THUMBPOSITION:
			newPos=(long) nPos; break;
		case SB_BOTTOM:
			newPos=yMax; break;
		case SB_TOP:
			newPos=0; break;
		case SB_ENDSCROLL:
			break;
		}
		if (newPos != yPos)
		{
			SCROLLINFO si;
			si.cbSize=sizeof(si);
			si.fMask=SIF_POS;
			si.nPos=newPos;
			SetScrollInfo(hWnd,SB_VERT,&si,TRUE);
			GetScrollInfo(hWnd,SB_VERT,&si);

			ScrollWindowEx(hWnd, 0, yPos-si.nPos, NULL,NULL,NULL,NULL,SW_ERASE);
			yPos=si.nPos;
			UpdateWindow(hWnd);
		}
	return TRUE;
}


BOOL  Window::HScroll(TMSG &Msg)
{
	if (!(Flags & W_SCROLLBARS)) return TRUE;

	int nScrollCode = LOWORD(Msg.wParam);   /* scroll bar value     */
	int nPos = LLHW(Msg);   /* scroll box position  */

	long newPos=xPos;
	switch (nScrollCode) {
		case SB_LINELEFT:
			newPos-=xStep; break;
		case SB_PAGELEFT:
			newPos-=xPageSize; break; // page??
		case SB_LINERIGHT:
			newPos+=xStep; break;
		case SB_PAGERIGHT:
			newPos+=xPageSize; break;
		case SB_THUMBTRACK:
		case SB_THUMBPOSITION:
			newPos=(long) nPos; break;
		case SB_LEFT:
			newPos=0; break;
		case SB_RIGHT:
			newPos=xMax; break;
		case SB_ENDSCROLL:
			break;
	}

	if (newPos != xPos)
	{
		SCROLLINFO si;
		si.cbSize=sizeof(si);
		si.fMask=SIF_POS;
		si.nPos=newPos;
		SetScrollInfo(hWnd,SB_HORZ,&si,TRUE);
		GetScrollInfo(hWnd,SB_HORZ,&si);

		ScrollWindowEx(hWnd, xPos-si.nPos, 0, NULL,NULL,NULL,NULL,SW_ERASE);
		xPos=si.nPos;
		UpdateWindow(hWnd);
	}

	return TRUE;
}

void Window::SetIcon(int id)
{
	Icon=id;
	if (hWnd && Icon>=0)
	{
		HICON I=LoadIcon(App::hInstance,MAKEINTRESOURCE(id));
		HICON OldIcon=(HICON) SetClassLong(hWnd,GCL_HICON,(LONG) I);
		if (OldIcon) DestroyIcon(OldIcon);
	}
}
